This repository contains code for the paper "MCMC computations for Bayesian mixture models using repulsive point processes" by Mario Beraha, Raffaele Argiento, Jesper M{\o}ller and Alessandra Guglielmi.

This is not a general-purpose statistical software, use at your own risk!

The code consists of a C++ library that computes MCMC posterior simulations for the models described in the paper, and is incapsulated in a 'python-like' package, for ease of use.

## Prerequisites

The code was developed and tested on an Ubuntu 18.04 machine and runs in Python3, make sure Python3 is installed. The following instructions lead to successful installation on Linux machines and should be identical for macOS. 
For Windows users: things should work as well modulo the fact that you should be able to install Python packages, have a valid and reasonably new C++ toolchain installed (gcc >= 7.5) and can successfully install and compile the following dependencies.

1. Protobuf: to install the protocol buffer library (C++ runtime library and protoc compiler) follow the [instructions](https://github.com/protocolbuffers/protobuf/tree/master/src). Make sure that the installed library is visible in the PATH (for instance, on Unix machines use `./configure --prefix=/usr`)

2. stan/math: we make extensive use of the awesome math library developed by the Stan team. Simply clone their repo (https://github.com/stan-dev/math) in a local directory and install it. An example of a real instantiation whenever the path to Stan Math is ~/stan-dev/math/:
```shell
make -j4 -f ~/stan-dev/math/make/standalone math-libs
make -f ~/stan-dev/math/make/standalone foo
```
Then set the environmental variable 'STAN_ROOT_DIR' to the path to 'math'.

3. pybind11
```shell
  pip3 install pybind11
```

4. 2to3
```shell
  sudo apt-get install 2to3
  sudo apt-get install python3-lib2to3
  sudo apt-get install python3-toolz
```

4. Python packages
```shell
  python3 -m pip install requirements.txt
```

## Installation
Installation is trivial on Linux systems and has been tested only on those.
```shell
  cd pp_mix
  make compile_protos
  make generate_pybind
```
and the package is ready to be used!

## Using the package
Here's a quick example
```python
import numpy as np
from pp_mix.interface import ConditionalMCMC, estimate_density_seq
from pp_mix.params_helper import make_default_strauss
import pp_mix.protos.py.params_pb2 as params_pb2

data = np.concatenate(
    [np.random.normal(100, loc=-5), np.random.normal(100, loc=5)])
    
# set prior parameters
pp_params = make_default_strauss(data)
prec_params = params_pb2.GammaParams(alpha=1.0, beta=1.0)
jump_params = params_pb2.GammaParams(alpha=1, beta=1)

# define sampler object
sampler = ConditionalMCMC(
    pp_params=pp_params, prec_params=prec_params,
    jump_params=gamma_jump_params)
    
# run MCMC chains
sampler.run(50000, 50000, 5, data)
xgrid = np.linspace(-10, 10, 1000)
dens_strauss = estimate_density_seq(sampler.chains, xgrid)
plt.plot(xgrid, np.mean(dens_strauss, axis=0))
```
To substitute the Strauss point process with the DPP, just use
```python
pp_params = params_pb2.DPPParams(nu=2.0, rho=3.0, N=50)
```
and use the rest of the code as before.


## Folder structure
The `data` folder contains the Teenager problematic behavior dataset (`add_health_extended.csv`) and the simulated datasets employed.

The source code is in the folder `pp_mix`. In particular, the file `interface.py` defines the classes `ConditionalMCMC` and `RJMCMC`, which are the main classes to call from Python: `ConditionalMCMC` is the Metropolis-within-Gibbs algorithm presented in our paper, `RJMCMC` is the split-merge reversible-jump competitor.

Under the hood, everything runs in C++, all the relevant files are contained in the `src` folder. The main files are `conditional_mcmc.hpp`, `conditional_mcmc_imp.hpp` and `conditional_mcmc.cpp`, where the C++ template class `ConditionalMCMC` is implemented. This class should not be instantiated directly, but one among the classes `MultivariateConditionalMCMC`, `UnivariateConditionalMCMC` and `BernoulliConditionalMCMC` should be used.
Different choices of the repulsive point process prior, the prior for the unnormalized weights and the precision can be passed to the constructor of the MCMC object.

The folder `src/point_process` contains the classes that represent the Strauss and determinantal point processes, they both inherit from the abstract base class `BasePP`, where some shared methods are implemented. Also, in this folder the `PerfectSampler` used in the exchange algorithm is implemented.
The folders `src/jumps` and `src/precs` implement the different choices of priors for the unnormalized weights (currently we support only the Gamma distribution) and the precision.

The folder `protos` collects the protocol buffer message definitions. These are used to represent the state of the MCMC algorithm, to pass these between C++ and Python and to serialize them and read them from finally. The model definition is done in Python, where the user uses the parameters proto messages to define the prior parameters, these are then passed to C++.


## Reproducing the results from the paper
First, run the R script `compare.R`: it contains instruction to run examples for Finite Mixture models and Dirichlet Process mixtures. Instructions to install the required packages are given inside the script itself.

After that, open one of the Jupyter notebooks and run it. Each notebook contains one of the simulation studies (or the real data analysis)


