library(AntMAN)
library(BNPmix)
library(coda)


fit_antman_multi <- function(data) {
  mcmc_params = AM_mcmc_parameters(niter=50000, burnin=25000, thin=5,
                                   output = c("CI","K","M","Tau","H","Q","S"))
  
  d = dim(data)[2]
  mixture_mvn_params = AM_mix_hyperparams_multinorm(mu0=rep(0, d), ka0=.5, nu0=4, Lam0=diag(d))
  components_prior = AM_mix_components_prior_pois (init=4, a=10, b=12)
  weights_prior = AM_mix_weights_prior_gamma(init=2, a=1, b=1)
  fit <- AM_mcmc_fit(
    y = as.matrix(data),
    mix_kernel_hyperparams = mixture_mvn_params,
    mix_components_prior = components_prior,
    mix_weight_prior = weights_prior,
    mcmc_parameters = mcmc_params)
  fit
}


fit_antman_univ <- function(data) {
  mixture_uni_params = AM_mix_hyperparams_uninorm(m0=0, k0=.1, nu0=1, sig02=1)
  mcmc_params = AM_mcmc_parameters(niter=50000, burnin=25000, thin=5,
                                   output = c("CI","K","M","Tau","H","Q","S"))
  components_prior = AM_mix_components_prior_pois (init=10, a=2, b=0.5)
  weights_prior = AM_mix_weights_prior_gamma(a=2, b=2)
  fit <- AM_mcmc_fit(
    y = data,
    mix_kernel_hyperparams = mixture_uni_params,
    mix_components_prior = components_prior,
    mix_weight_prior = weights_prior,
    mcmc_parameters = mcmc_params)
  fit
}



fit_bnpmix <- function(data, grid) {
  est_model <- PYdensity(y = data, mcmc = list(niter = 10000, nburn = 5000, method="MAR", hyper=F),
                         prior = list(k0 = 0.01, strength = 5, discount = 0.0),
                         output = list(grid = grid))
  est_model
}

fit_bnpmix_noest <- function(data) {
  est_model <- PYdensity(y = data, mcmc = list(niter = 10000, nburn = 5000, method="ICS", hyper=F),
                         prior = list(k0 = 0.01, strength = 15, discount = 0.0))
  est_model
}


##################################
### MIXTURE OF T AND SKEW-NORMAL #
##################################

data_files = c(
  "data/sim_t_skew_dim1.txt",
  "data/sim_t_skew_dim5.txt")

# Univariate Case
data = read.csv(data_files[1], sep=" ", header = F)$V1
fit <- fit_antman_univ(data)
write.table(fit$K, "data/sim_t_skew_dim1_Kfm.txt", sep=" ", row.names=F, col.names=F)
write.table(fit$CI, "data/sim_t_skew_dim1_c_fm.txt", sep=" ", row.names=F, col.names=F)

grid <- seq(-10, 10, length.out = 100)
est_model <- fit_bnpmix_noest(data)
nclus = apply(est_model$clust + 1, 1, max)
write.table(nclus, "data/sim_t_skew_dim1_Kbnpmix.txt", sep=" ", row.names=F, col.names=F)
write.table(est_model$clust, "data/sim_t_skew_dim1_Cbnpmix.txt", sep=" ", row.names=F, col.names=F)
table(nclus)

# Multivariate Cases
data = as.matrix(read.csv(data_files[2], sep=" ", header = F))
fit <- fit_antman_multi(data)
write.table(fit$K, "data/sim_t_skew_dim2_Kraf.txt", sep=" ", row.names=F, col.names=F)
write.table(fit$CI, "data/sim_t_skew_dim5_c_fm.txt", sep=" ", row.names=F, col.names=F)

est_model <- fit_bnpmix_noest(data)
nclus = apply(est_model$clust, 1, max)
write.table(nclus, "data/sim_t_skew_dim2_Kbnpmix.txt", sep=" ", row.names=F, col.names=F)



######################
### MILLER DUNSON ####
######################

data = read.csv("data/data_univ_miller_dunson.txt", sep=" ", header = F)$V1
hist(data)
fit <- fit_antman_univ(data)

write.table(fit$M, "data/data_univ_miller_dunson_M_fm.txt", sep=" ", row.names=F, col.names=F)
write.table(fit$K, "data/data_univ_miller_dunson_K_fm.txt", sep=" ", row.names=F, col.names=F)

grid <- seq(-10, 10, length.out=200)
est_model <- fit_bnpmix(data, grid)

write.table(est_model$density, "data/data_univ_miller_dunson_dens_bnpmix.txt", sep=" ", row.names=F, col.names=F)
nclus = apply(est_model$clust, 1, max)
hist(nclus)
write.table(nclus, "data/data_univ_miller_dunson_K_bnpmix.txt", sep=" ", row.names=F, col.names=F)
mean(nclus)
